@@ -24,7 +24,7 @@
 /**
  * SECTION:gstclock
  * @short_description: Abstract class for global clocks
- * @see_also: #GstSystemClock
+ * @see_also: #GstSystemClock, #GstPipeline
  *
  * GStreamer uses a global clock to synchronize the plugins in a pipeline.
  * Different clock implementations are possible by implementing this abstract
@@ -35,10 +35,11 @@
  * clock implementation but time is always expressed in nanoseconds. Since the
  * baseline of the clock is undefined, the clock time returned is not
  * meaningful in itself, what matters are the deltas between two clock times.
+ * The time returned by a clock is called the absolute time.
  *
  * The pipeline uses the clock to calculate the stream time. Usually all
  * renderers synchronize to the global clock using the buffer timestamps, the
- * newsegment events and the element's base time.
+ * newsegment events and the element's base time, see #GstPipeline.
  *
  * A clock implementation can support periodic and single shot clock
  * notifications both synchronous and asynchronous.
@@ -85,7 +86,7 @@
  * slaved to another #GstClock with the gst_clock_set_master(). The clock will
  * then automatically be synchronized to this master clock by repeadedly
  * sampling the master clock and the slave clock and recalibrating the slave
- * clock with gst_clock_set_calibration(). This feature is mostly usefull for
+ * clock with gst_clock_set_calibration(). This feature is mostly useful for
  * plugins that have an internal clock but must operate with another clock
  * selected by the #GstPipeline.  They can track the offset and rate difference
  * of their internal clock relative to the master clock by using the
@@ -98,12 +99,12 @@
  * "window-threshold" defines the minimum number of samples before the 
  * calibration is performed.
  *
- * Last reviewed on 2006-03-09 (0.10.4)
+ * Last reviewed on 2006-08-11 (0.10.10)
  */
 
-#include <time.h>
 
 #include "gst_private.h"
+#include <time.h>
 
 #include "gstclock.h"
 #include "gstinfo.h"
@@ -115,6 +116,14 @@
 static GstAllocTrace *_gst_clock_entry_trace;
 #endif
 
+#if GLIB_CHECK_VERSION (2, 10, 0)
+#define ALLOC_ENTRY()     g_slice_new (GstClockEntry)
+#define FREE_ENTRY(entry) g_slice_free (GstClockEntry, entry)
+#else
+#define ALLOC_ENTRY()     g_new (GstClockEntry, 1)
+#define FREE_ENTRY(entry) g_free (entry)
+#endif
+
 /* #define DEBUGGING_ENABLED */
 
 #define DEFAULT_STATS                   FALSE
@@ -153,9 +162,7 @@ gst_clock_entry_new (GstClock * clock, GstClockTime time,
 {
   GstClockEntry *entry;
 
-  /* FIXME, use g_slice, we do this a lot and potentially from
-   * different threads. */
-  entry = g_new0 (GstClockEntry, 1);
+  entry = ALLOC_ENTRY ();
 #ifndef GST_DISABLE_TRACE
   gst_alloc_trace_new (_gst_clock_entry_trace, entry);
 #endif
@@ -164,10 +171,12 @@ gst_clock_entry_new (GstClock * clock, GstClockTime time,
 
   gst_atomic_int_set (&entry->refcount, 1);
   entry->clock = clock;
+  entry->type = type;
   entry->time = time;
   entry->interval = interval;
-  entry->type = type;
   entry->status = GST_CLOCK_BUSY;
+  entry->func = NULL;
+  entry->user_data = NULL;
 
   return (GstClockID) entry;
 }
@@ -202,7 +211,7 @@ _gst_clock_id_free (GstClockID id)
 #ifndef GST_DISABLE_TRACE
   gst_alloc_trace_free (_gst_clock_entry_trace, id);
 #endif
-  g_free (id);
+  FREE_ENTRY (id);
 }
 
 /**
@@ -272,6 +281,7 @@ gst_clock_new_periodic_id (GstClock * clock, GstClockTime start_time,
   g_return_val_if_fail (GST_IS_CLOCK (clock), NULL);
   g_return_val_if_fail (GST_CLOCK_TIME_IS_VALID (start_time), NULL);
   g_return_val_if_fail (interval != 0, NULL);
+  g_return_val_if_fail (GST_CLOCK_TIME_IS_VALID (interval), NULL);
 
   return gst_clock_entry_new (clock,
       start_time, interval, GST_CLOCK_ENTRY_PERIODIC);
@@ -324,7 +334,6 @@ gst_clock_id_get_time (GstClockID id)
   return GST_CLOCK_ENTRY_TIME ((GstClockEntry *) id);
 }
 
-
 /**
  * gst_clock_id_wait
  * @id: The #GstClockID to wait on
@@ -338,9 +347,11 @@ gst_clock_id_get_time (GstClockID id)
  * If the @jitter argument is not NULL and this function returns #GST_CLOCK_OK
  * or #GST_CLOCK_EARLY, it will contain the difference
  * against the clock and the time of @id when this method was
- * called. Negative values means @id was scheduled too late (and this
- * function will return #GST_CLOCK_EARLY). Positive values indicate how
- * early @id was scheduled. 
+ * called. 
+ * Positive values indicate how late @id was relative to the clock
+ * (in which case this function will return #GST_CLOCK_EARLY). 
+ * Negative values indicate how much time was spent waiting on the clock 
+ * before this function returned.
  *
  * Returns: the result of the blocking wait. #GST_CLOCK_EARLY will be returned
  * if the current clock time is past the time of @id, #GST_CLOCK_OK if 
@@ -375,25 +386,35 @@ gst_clock_id_wait (GstClockID id, GstClockTimeDiff * jitter)
 
   cclass = GST_CLOCK_GET_CLASS (clock);
 
-  if (G_UNLIKELY (cclass->wait == NULL))
-    goto not_supported;
-
   GST_CAT_DEBUG_OBJECT (GST_CAT_CLOCK, clock, "waiting on clock entry %p", id);
 
-  /* jitter is the diff against the clock when this entry is scheduled */
-  if (jitter) {
-    GstClockTime now = gst_clock_get_time (clock);
+  /* if we have a wait_jitter function, use that */
+  if (G_LIKELY (cclass->wait_jitter)) {
+    res = cclass->wait_jitter (clock, entry, jitter);
+  } else {
+    /* check if we have a simple _wait function otherwise. The function without
+     * the jitter arg is less optimal as we need to do an additional _get_time()
+     * which is not atomic with the _wait() and a typical _wait() function does
+     * yet another _get_time() anyway. */
+    if (G_UNLIKELY (cclass->wait == NULL))
+      goto not_supported;
 
-    *jitter = GST_CLOCK_DIFF (requested, now);
+    if (jitter) {
+      GstClockTime now = gst_clock_get_time (clock);
+
+      /* jitter is the diff against the clock when this entry is scheduled. Negative
+       * values mean that the entry was in time, a positive value means that the
+       * entry was too late. */
+      *jitter = GST_CLOCK_DIFF (requested, now);
+    }
+    res = cclass->wait (clock, entry);
   }
-  res = cclass->wait (clock, entry);
 
   GST_CAT_DEBUG_OBJECT (GST_CAT_CLOCK, clock,
       "done waiting entry %p, res: %d", id, res);
 
-  if (entry->type == GST_CLOCK_ENTRY_PERIODIC) {
-    entry->time += entry->interval;
-  }
+  if (entry->type == GST_CLOCK_ENTRY_PERIODIC)
+    entry->time = requested + entry->interval;
 
   if (clock->stats)
     gst_clock_update_stats (clock);
@@ -558,10 +579,10 @@ gst_clock_class_init (GstClockClass * klass)
   GObjectClass *gobject_class;
   GstObjectClass *gstobject_class;
 
-  gobject_class = (GObjectClass *) klass;
-  gstobject_class = (GstObjectClass *) klass;
+  gobject_class = G_OBJECT_CLASS (klass);
+  gstobject_class = GST_OBJECT_CLASS (klass);
 
-  parent_class = g_type_class_ref (GST_TYPE_OBJECT);
+  parent_class = g_type_class_peek_parent (klass);
 
   if (!g_thread_supported ())
     g_thread_init (NULL);
@@ -576,19 +597,19 @@ gst_clock_class_init (GstClockClass * klass)
   gobject_class->set_property = GST_DEBUG_FUNCPTR (gst_clock_set_property);
   gobject_class->get_property = GST_DEBUG_FUNCPTR (gst_clock_get_property);
 
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_STATS,
-      g_param_spec_boolean ("stats", "Stats", "Enable clock stats",
-          DEFAULT_STATS, G_PARAM_READWRITE));
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_WINDOW_SIZE,
+  g_object_class_install_property (gobject_class, PROP_STATS,
+      g_param_spec_boolean ("stats", "Stats",
+          "Enable clock stats (unimplemented)", DEFAULT_STATS,
+          G_PARAM_READWRITE));
+  g_object_class_install_property (gobject_class, PROP_WINDOW_SIZE,
       g_param_spec_int ("window-size", "Window size",
           "The size of the window used to calculate rate and offset", 2, 1024,
           DEFAULT_WINDOW_SIZE, G_PARAM_READWRITE));
-  g_object_class_install_property (G_OBJECT_CLASS (klass),
-      PROP_WINDOW_THRESHOLD, g_param_spec_int ("window-threshold",
-          "Window threshold",
+  g_object_class_install_property (gobject_class, PROP_WINDOW_THRESHOLD,
+      g_param_spec_int ("window-threshold", "Window threshold",
           "The threshold to start calculating rate and offset", 2, 1024,
           DEFAULT_WINDOW_THRESHOLD, G_PARAM_READWRITE));
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_TIMEOUT,
+  g_object_class_install_property (gobject_class, PROP_TIMEOUT,
       g_param_spec_uint64 ("timeout", "Timeout",
           "The amount of time, in nanoseconds, to sample master and slave clocks",
           0, G_MAXUINT64, DEFAULT_TIMEOUT, G_PARAM_READWRITE));
@@ -620,9 +641,11 @@ static void
 gst_clock_dispose (GObject * object)
 {
   GstClock *clock = GST_CLOCK (object);
+  GstClock **master_p;
 
   GST_OBJECT_LOCK (clock);
-  gst_object_replace ((GstObject **) & clock->master, NULL);
+  master_p = &clock->master;
+  gst_object_replace ((GstObject **) master_p, NULL);
   GST_OBJECT_UNLOCK (clock);
 
   G_OBJECT_CLASS (parent_class)->dispose (object);
@@ -654,7 +677,11 @@ gst_clock_finalize (GObject * object)
  * @clock: a #GstClock
  * @resolution: The resolution to set
  *
- * Set the accuracy of the clock.
+ * Set the accuracy of the clock. Some clocks have the possibility to operate
+ * with different accuracy at the expense of more resource usage. There is
+ * normally no need to change the default resolution of a clock. The resolution
+ * of a clock can only be changed if the clock has the
+ * #GST_CLOCK_FLAG_CAN_SET_RESOLUTION flag set.
  *
  * Returns: the new resolution of the clock.
  */
@@ -679,7 +706,8 @@ gst_clock_set_resolution (GstClock * clock, GstClockTime resolution)
  * gst_clock_get_resolution
  * @clock: a #GstClock
  *
- * Get the accuracy of the clock.
+ * Get the accuracy of the clock. The accuracy of the clock is the granularity
+ * of the values returned by gst_clock_get_time().
  *
  * Returns: the resolution of the clock in units of #GstClockTime.
  *
@@ -936,6 +964,8 @@ gst_clock_slave_callback (GstClock * master, GstClockTime time,
 gboolean
 gst_clock_set_master (GstClock * clock, GstClock * master)
 {
+  GstClock **master_p;
+
   g_return_val_if_fail (GST_IS_CLOCK (clock), FALSE);
   g_return_val_if_fail (master != clock, FALSE);
 
@@ -945,7 +975,8 @@ gst_clock_set_master (GstClock * clock, GstClock * master)
     goto not_supported;
 
   GST_DEBUG_OBJECT (clock, "slaving to master clock %p", master);
-  gst_object_replace ((GstObject **) & clock->master, (GstObject *) master);
+  master_p = &clock->master;
+  gst_object_replace ((GstObject **) master_p, (GstObject *) master);
   GST_OBJECT_UNLOCK (clock);
 
   GST_CLOCK_SLAVE_LOCK (clock);
@@ -1113,6 +1144,7 @@ do_linear_regression (GstClock * clock, GstClockTime * m_num,
 
 invalid:
   {
+    GST_DEBUG_OBJECT (clock, "sxx == 0, regression failed");
     return FALSE;
   }
 }
@@ -1129,6 +1161,12 @@ invalid:
  * are available, a linear regression algorithm is run on the
  * observations and @clock is recalibrated.
  *
+ * If this functions returns %TRUE, @r_squared will contain the 
+ * correlation coefficient of the interpollation. A value of 1.0
+ * means a perfect regression was performed. This value can
+ * be used to control the sampling frequency of the master and slave
+ * clocks.
+ *
  * Returns: TRUE if enough observations were added to run the 
  * regression algorithm.
  *
