@@ -28,6 +28,63 @@
  *
  * One registry holds the metadata of a set of plugins.
  * All registries build the #GstRegistryPool.
+ * 
+ * <emphasis role="bold">Design:</emphasis>
+ *
+ * The #GstRegistry object is a list of plugins and some functions for dealing
+ * with them. #GstPlugins are matched 1-1 with a file on disk, and may or may
+ * not be loaded at a given time. There may be multiple #GstRegistry objects,
+ * but the "default registry" is the only object that has any meaning to the
+ * core.
+ *
+ * The registry.xml file is actually a cache of plugin information. This is
+ * unlike versions prior to 0.10, where the registry file was the primary source
+ * of plugin information, and was created by the gst-register command.
+ *
+ * The primary source, at all times, of plugin information is each plugin file
+ * itself. Thus, if an application wants information about a particular plugin,
+ * or wants to search for a feature that satisfies given criteria, the primary
+ * means of doing so is to load every plugin and look at the resulting
+ * information that is gathered in the default registry. Clearly, this is a time
+ * consuming process, so we cache information in the registry.xml file.
+ *
+ * On startup, plugins are searched for in the plugin search path. This path can
+ * be set directly using the %GST_PLUGIN_PATH environment variable. The registry
+ * file is loaded from ~/.gstreamer-$GST_MAJORMINOR/registry-$ARCH.xml or the
+ * file listed in the %GST_REGISTRY env var. The only reason to change the
+ * registry location is for testing.
+ *
+ * For each plugin that is found in the plugin search path, there could be 3
+ * possibilities for cached information:
+ * <itemizedlist>
+ *   <listitem>
+ *     <para>the cache may not contain information about a given file.</para>
+ *   </listitem>
+ *   <listitem>
+ *     <para>the cache may have stale information.</para>
+ *   </listitem>
+ *   <listitem>
+ *     <para>the cache may have current information.</para>
+ *   </listitem>
+ * </itemizedlist>
+ *
+ * In the first two cases, the plugin is loaded and the cache updated. In
+ * addition to these cases, the cache may have entries for plugins that are not
+ * relevant to the current process. These are marked as not available to the
+ * current process. If the cache is updated for whatever reason, it is marked
+ * dirty.
+ *
+ * A dirty cache is written out at the end of initialization. Each entry is
+ * checked to make sure the information is minimally valid. If not, the entry is
+ * simply dropped.
+ *
+ * <emphasis role="bold">Implementation notes:</emphasis>
+ *
+ * The "cache" and "default registry" are different concepts and can represent
+ * different sets of plugins. For various reasons, at init time, the cache is
+ * stored in the default registry, and plugins not relevant to the current
+ * process are marked with the %GST_PLUGIN_FLAG_CACHED bit. These plugins are
+ * removed at the end of intitialization.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -52,65 +109,11 @@
 
 #define GST_CAT_DEFAULT GST_CAT_REGISTRY
 
-/* the one instance of the default registry */
+/* the one instance of the default registry and the mutex protecting the 
+ * variable. */
+static GStaticMutex _gst_registry_mutex = G_STATIC_MUTEX_INIT;
 static GstRegistry *_gst_registry_default = NULL;
 
-/*
- * Design:
- *
- * The GstRegistry object is a list of plugins and some functions
- * for dealing with them.  Plugins are matched 1-1 with a file on
- * disk, and may or may not be loaded at a given time.  There may
- * be multiple GstRegistry objects, but the "default registry" is
- * the only object that has any meaning to the core.
- *
- * The registry.xml file in 0.9 is actually a cache of plugin
- * information.  This is unlike previous versions, where the registry
- * file was the primary source of plugin information, and was created
- * by the gst-register command.
- *
- * In 0.9, the primary source, at all times, of plugin information
- * is each plugin file itself.  Thus, if an application wants
- * information about a particular plugin, or wants to search for
- * a feature that satisfies given criteria, the primary means of
- * doing so is to load every plugin and look at the resulting
- * information that is gathered in the default registry.  Clearly,
- * this is a time consuming process, so we cache information in
- * the registry.xml file.
- *
- * On startup, plugins are searched for in the plugin search path.
- * This path can be set directly using the GST_PLUGIN_PATH
- * environment variable.  The registry file is loaded from
- * ~/.gstreamer-0.9/registry.xml or the file listed in the
- * GST_REGISTRY env var.  The only reason to change the registry
- * location is for testing.
- *
- * For each plugin that is found in the plugin search path, there
- * could be 3 possibilities for cached information:
- *  - the cache may not contain information about a given file.
- *  - the cache may have stale information.
- *  - the cache may have current information.
- * In the first two cases, the plugin is loaded and the cache
- * updated.  In addition to these cases, the cache may have entries
- * for plugins that are not relevant to the current process.  These
- * are marked as not available to the current process.  If the
- * cache is updated for whatever reason, it is marked dirty.
- *
- * A dirty cache is written out at the end of initialization.  Each
- * entry is checked to make sure the information is minimally valid.
- * If not, the entry is simply dropped.
- *
- * Implementation notes:
- *
- * The "cache" and "default registry" are different concepts and
- * can represent different sets of plugins.  For various reasons,
- * at init time, the cache is stored in the default registry, and
- * plugins not relevant to the current process are marked with the
- * GST_PLUGIN_FLAG_CACHED bit.  These plugins are removed at the
- * end of intitialization.
- *
- */
-
 /* Element signals and args */
 enum
 {
@@ -140,12 +143,29 @@ gst_registry_class_init (GstRegistryClass * klass)
 
   gobject_class = (GObjectClass *) klass;
 
-  parent_class = g_type_class_ref (GST_TYPE_OBJECT);
+  parent_class = g_type_class_peek_parent (klass);
 
+  /**
+   * GstRegistry::plugin-added:
+   * @registry: the registry that emitted the signal
+   * @plugin: the plugin that has been added
+   *
+   * Signals that a plugin has been added to the registry (possibly
+   * replacing a previously-added one by the same name)
+   */
   gst_registry_signals[PLUGIN_ADDED] =
       g_signal_new ("plugin-added", G_TYPE_FROM_CLASS (klass),
       G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (GstRegistryClass, plugin_added), NULL,
       NULL, gst_marshal_VOID__POINTER, G_TYPE_NONE, 1, G_TYPE_POINTER);
+
+  /**
+   * GstRegistry::feature-added:
+   * @registry: the registry that emitted the signal
+   * @feature: the feature that has been added
+   *
+   * Signals that a feature has been added to the registry (possibly
+   * replacing a previously-added one by the same name)
+   */
   gst_registry_signals[FEATURE_ADDED] =
       g_signal_new ("feature-added", G_TYPE_FROM_CLASS (klass),
       G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (GstRegistryClass, feature_added),
@@ -175,9 +195,9 @@ gst_registry_finalize (GObject * object)
     GstPlugin *plugin = p->data;
 
     if (plugin) {
-      GST_DEBUG_OBJECT (registry, "removing plugin %s",
+      GST_LOG_OBJECT (registry, "removing plugin %s",
           gst_plugin_get_name (plugin));
-      gst_registry_remove_plugin (registry, plugin);
+      gst_object_unref (plugin);
     }
     p = g_list_next (p);
   }
@@ -191,7 +211,9 @@ gst_registry_finalize (GObject * object)
     GstPluginFeature *feature = f->data;
 
     if (feature) {
-      gst_registry_remove_feature (registry, feature);
+      GST_LOG_OBJECT (registry, "removing feature %p (%s)",
+          feature, gst_plugin_feature_get_name (feature));
+      gst_object_unref (feature);
     }
     f = g_list_next (f);
   }
@@ -212,10 +234,18 @@ gst_registry_finalize (GObject * object)
 GstRegistry *
 gst_registry_get_default (void)
 {
-  if (!_gst_registry_default) {
+  GstRegistry *registry;
+
+  g_static_mutex_lock (&_gst_registry_mutex);
+  if (G_UNLIKELY (!_gst_registry_default)) {
     _gst_registry_default = g_object_new (GST_TYPE_REGISTRY, NULL);
+    gst_object_ref (GST_OBJECT_CAST (_gst_registry_default));
+    gst_object_sink (GST_OBJECT_CAST (_gst_registry_default));
   }
-  return _gst_registry_default;
+  registry = _gst_registry_default;
+  g_static_mutex_unlock (&_gst_registry_mutex);
+
+  return registry;
 }
 
 /**
@@ -304,10 +334,11 @@ gst_registry_add_plugin (GstRegistry * registry, GstPlugin * plugin)
   GstPlugin *existing_plugin;
 
   g_return_val_if_fail (GST_IS_REGISTRY (registry), FALSE);
+  g_return_val_if_fail (GST_IS_PLUGIN (plugin), FALSE);
 
   GST_OBJECT_LOCK (registry);
   existing_plugin = gst_registry_lookup_locked (registry, plugin->filename);
-  if (existing_plugin) {
+  if (G_UNLIKELY (existing_plugin)) {
     GST_DEBUG_OBJECT (registry,
         "Replacing existing plugin %p with new plugin %p for filename \"%s\"",
         existing_plugin, plugin, GST_STR_NULL (plugin->filename));
@@ -324,7 +355,7 @@ gst_registry_add_plugin (GstRegistry * registry, GstPlugin * plugin)
   gst_object_sink (plugin);
   GST_OBJECT_UNLOCK (registry);
 
-  GST_DEBUG_OBJECT (registry, "emitting plugin-added for filename \"%s\"",
+  GST_LOG_OBJECT (registry, "emitting plugin-added for filename \"%s\"",
       GST_STR_NULL (plugin->filename));
   g_signal_emit (G_OBJECT (registry), gst_registry_signals[PLUGIN_ADDED], 0,
       plugin);
@@ -332,6 +363,33 @@ gst_registry_add_plugin (GstRegistry * registry, GstPlugin * plugin)
   return TRUE;
 }
 
+static void
+gst_registry_remove_features_for_plugin_unlocked (GstRegistry * registry,
+    GstPlugin * plugin)
+{
+  GList *f;
+
+  g_return_if_fail (GST_IS_REGISTRY (registry));
+  g_return_if_fail (GST_IS_PLUGIN (plugin));
+
+  /* Remove all features for this plugin */
+  f = registry->features;
+  while (f != NULL) {
+    GList *next = g_list_next (f);
+    GstPluginFeature *feature = f->data;
+
+    if (feature && !strcmp (feature->plugin_name, gst_plugin_get_name (plugin))) {
+      GST_DEBUG_OBJECT (registry, "removing feature %p (%s) for plugin %s",
+          feature, gst_plugin_feature_get_name (feature),
+          gst_plugin_get_name (plugin));
+
+      registry->features = g_list_delete_link (registry->features, f);
+      gst_object_unref (feature);
+    }
+    f = next;
+  }
+}
+
 /**
  * gst_registry_remove_plugin:
  * @registry: the registry to remove the plugin from
@@ -345,9 +403,14 @@ void
 gst_registry_remove_plugin (GstRegistry * registry, GstPlugin * plugin)
 {
   g_return_if_fail (GST_IS_REGISTRY (registry));
+  g_return_if_fail (GST_IS_PLUGIN (plugin));
+
+  GST_DEBUG_OBJECT (registry, "removing plugin %p (%s)",
+      plugin, gst_plugin_get_name (plugin));
 
   GST_OBJECT_LOCK (registry);
   registry->plugins = g_list_remove (registry->plugins, plugin);
+  gst_registry_remove_features_for_plugin_unlocked (registry, plugin);
   GST_OBJECT_UNLOCK (registry);
   gst_object_unref (plugin);
 }
@@ -377,7 +440,7 @@ gst_registry_add_feature (GstRegistry * registry, GstPluginFeature * feature)
   GST_OBJECT_LOCK (registry);
   existing_feature = gst_registry_lookup_feature_locked (registry,
       feature->name);
-  if (existing_feature) {
+  if (G_UNLIKELY (existing_feature)) {
     GST_DEBUG_OBJECT (registry, "Replacing existing feature %p (%s)",
         existing_feature, feature->name);
     registry->features = g_list_remove (registry->features, existing_feature);
@@ -392,7 +455,7 @@ gst_registry_add_feature (GstRegistry * registry, GstPluginFeature * feature)
   gst_object_sink (feature);
   GST_OBJECT_UNLOCK (registry);
 
-  GST_DEBUG_OBJECT (registry, "emitting feature-added for %s", feature->name);
+  GST_LOG_OBJECT (registry, "emitting feature-added for %s", feature->name);
   g_signal_emit (G_OBJECT (registry), gst_registry_signals[FEATURE_ADDED], 0,
       feature);
 
@@ -412,6 +475,8 @@ void
 gst_registry_remove_feature (GstRegistry * registry, GstPluginFeature * feature)
 {
   g_return_if_fail (GST_IS_REGISTRY (registry));
+  g_return_if_fail (GST_IS_PLUGIN_FEATURE (feature));
+
   GST_DEBUG_OBJECT (registry, "removing feature %p (%s)",
       feature, gst_plugin_feature_get_name (feature));
 
@@ -451,7 +516,7 @@ gst_registry_plugin_filter (GstRegistry * registry,
   list = gst_filter_run (registry->plugins, (GstFilterFunc) filter, first,
       user_data);
   for (g = list; g; g = g->next) {
-    gst_object_ref (GST_PLUGIN (g->data));
+    gst_object_ref (GST_PLUGIN_CAST (g->data));
   }
   GST_OBJECT_UNLOCK (registry);
 
@@ -487,7 +552,7 @@ gst_registry_feature_filter (GstRegistry * registry,
   list = gst_filter_run (registry->features, (GstFilterFunc) filter, first,
       user_data);
   for (g = list; g; g = g->next) {
-    gst_object_ref (GST_PLUGIN_FEATURE (g->data));
+    gst_object_ref (GST_PLUGIN_FEATURE_CAST (g->data));
   }
   GST_OBJECT_UNLOCK (registry);
 
@@ -519,7 +584,7 @@ gst_registry_find_plugin (GstRegistry * registry, const gchar * name)
   walk = gst_registry_plugin_filter (registry,
       (GstPluginFilter) gst_plugin_name_filter, TRUE, (gpointer) name);
   if (walk) {
-    result = GST_PLUGIN (walk->data);
+    result = GST_PLUGIN_CAST (walk->data);
 
     gst_object_ref (result);
     gst_plugin_list_free (walk);
@@ -551,6 +616,7 @@ gst_registry_find_feature (GstRegistry * registry, const gchar * name,
 
   g_return_val_if_fail (GST_IS_REGISTRY (registry), NULL);
   g_return_val_if_fail (name != NULL, NULL);
+  g_return_val_if_fail (g_type_is_a (type, GST_TYPE_PLUGIN_FEATURE), NULL);
 
   data.name = name;
   data.type = type;
@@ -560,7 +626,7 @@ gst_registry_find_feature (GstRegistry * registry, const gchar * name,
       TRUE, &data);
 
   if (walk) {
-    feature = GST_PLUGIN_FEATURE (walk->data);
+    feature = GST_PLUGIN_FEATURE_CAST (walk->data);
 
     gst_object_ref (feature);
     gst_plugin_feature_list_free (walk);
@@ -586,6 +652,9 @@ gst_registry_get_feature_list (GstRegistry * registry, GType type)
 {
   GstTypeNameData data;
 
+  g_return_val_if_fail (GST_IS_REGISTRY (registry), NULL);
+  g_return_val_if_fail (g_type_is_a (type, GST_TYPE_PLUGIN_FEATURE), NULL);
+
   data.type = type;
   data.name = NULL;
 
@@ -611,10 +680,12 @@ gst_registry_get_plugin_list (GstRegistry * registry)
   GList *list;
   GList *g;
 
+  g_return_val_if_fail (GST_IS_REGISTRY (registry), NULL);
+
   GST_OBJECT_LOCK (registry);
   list = g_list_copy (registry->plugins);
   for (g = list; g; g = g->next) {
-    gst_object_ref (GST_PLUGIN (g->data));
+    gst_object_ref (GST_PLUGIN_CAST (g->data));
   }
   GST_OBJECT_UNLOCK (registry);
 
@@ -627,11 +698,12 @@ gst_registry_lookup_feature_locked (GstRegistry * registry, const char *name)
   GList *g;
   GstPluginFeature *feature;
 
-  if (name == NULL)
+  if (G_UNLIKELY (name == NULL))
     return NULL;
 
+  /* FIXME: use GTree speed up lookups */
   for (g = registry->features; g; g = g_list_next (g)) {
-    feature = GST_PLUGIN_FEATURE (g->data);
+    feature = GST_PLUGIN_FEATURE_CAST (g->data);
     if (feature->name && strcmp (name, feature->name) == 0) {
       return feature;
     }
@@ -657,6 +729,9 @@ gst_registry_lookup_feature (GstRegistry * registry, const char *name)
 {
   GstPluginFeature *feature;
 
+  g_return_val_if_fail (GST_IS_REGISTRY (registry), NULL);
+  g_return_val_if_fail (name != NULL, NULL);
+
   GST_OBJECT_LOCK (registry);
   feature = gst_registry_lookup_feature_locked (registry, name);
   if (feature)
@@ -673,12 +748,13 @@ gst_registry_lookup_locked (GstRegistry * registry, const char *filename)
   GstPlugin *plugin;
   gchar *basename;
 
-  if (filename == NULL)
+  if (G_UNLIKELY (filename == NULL))
     return NULL;
 
   basename = g_path_get_basename (filename);
+  /* FIXME: use GTree speed up lookups */
   for (g = registry->plugins; g; g = g_list_next (g)) {
-    plugin = GST_PLUGIN (g->data);
+    plugin = GST_PLUGIN_CAST (g->data);
     if (plugin->basename && strcmp (basename, plugin->basename) == 0) {
       g_free (basename);
       return plugin;
@@ -705,6 +781,9 @@ gst_registry_lookup (GstRegistry * registry, const char *filename)
 {
   GstPlugin *plugin;
 
+  g_return_val_if_fail (GST_IS_REGISTRY (registry), NULL);
+  g_return_val_if_fail (filename != NULL, NULL);
+
   GST_OBJECT_LOCK (registry);
   plugin = gst_registry_lookup_locked (registry, filename);
   if (plugin)
@@ -714,7 +793,7 @@ gst_registry_lookup (GstRegistry * registry, const char *filename)
   return plugin;
 }
 
-static void
+static gboolean
 gst_registry_scan_path_level (GstRegistry * registry, const gchar * path,
     int level)
 {
@@ -723,10 +802,11 @@ gst_registry_scan_path_level (GstRegistry * registry, const gchar * path,
   gchar *filename;
   GstPlugin *plugin;
   GstPlugin *newplugin;
+  gboolean changed = FALSE;
 
   dir = g_dir_open (path, 0, NULL);
   if (!dir)
-    return;
+    return FALSE;
 
   while ((dirent = g_dir_read_name (dir))) {
     filename = g_strjoin ("/", path, dirent, NULL);
@@ -736,7 +816,7 @@ gst_registry_scan_path_level (GstRegistry * registry, const gchar * path,
     if (g_file_test (filename, G_FILE_TEST_IS_DIR)) {
       if (level > 0) {
         GST_LOG_OBJECT (registry, "found directory, recursing");
-        gst_registry_scan_path_level (registry, filename, level - 1);
+        changed |= gst_registry_scan_path_level (registry, filename, level - 1);
       } else {
         GST_LOG_OBJECT (registry,
             "found directory, but recursion level is too deep");
@@ -750,6 +830,7 @@ gst_registry_scan_path_level (GstRegistry * registry, const gchar * path,
       continue;
     }
     if (!g_str_has_suffix (filename, ".so") &&
+        !g_str_has_suffix (filename, ".sl") &&
         !g_str_has_suffix (filename, ".dll") &&
         !g_str_has_suffix (filename, ".dynlib")) {
       GST_LOG_OBJECT (registry,
@@ -765,8 +846,10 @@ gst_registry_scan_path_level (GstRegistry * registry, const gchar * path,
       struct stat file_status;
 
       if (stat (filename, &file_status)) {
-        /* FIXME remove from cache */
+        /* Plugin will be removed from cache after the scan completes if it
+         * is still marked 'cached' */
         g_free (filename);
+        gst_object_unref (plugin);
         continue;
       }
       if (plugin->registered) {
@@ -779,16 +862,24 @@ gst_registry_scan_path_level (GstRegistry * registry, const gchar * path,
       }
       if (plugin->file_mtime == file_status.st_mtime &&
           plugin->file_size == file_status.st_size) {
-        GST_DEBUG_OBJECT (registry, "file %s cached", filename);
+        GST_LOG_OBJECT (registry, "file %s cached", filename);
         plugin->flags &= ~GST_PLUGIN_FLAG_CACHED;
-        GST_DEBUG_OBJECT (registry, "marking plugin %p as registered", plugin);
+        GST_LOG_OBJECT (registry, "marking plugin %p as registered as %s",
+            plugin, filename);
         plugin->registered = TRUE;
+        /* Update the file path on which we've seen this cached plugin 
+         * to ensure the registry cache will reflect up to date information */
+        if (strcmp (plugin->filename, filename) != 0) {
+          g_free (plugin->filename);
+          plugin->filename = g_strdup (filename);
+          changed = TRUE;
+        }
       } else {
         GST_INFO_OBJECT (registry, "cached info for %s is stale", filename);
         GST_DEBUG_OBJECT (registry, "mtime %ld != %ld or size %"
-            G_GSIZE_FORMAT " != %"
-            G_GSIZE_FORMAT, plugin->file_mtime, file_status.st_mtime,
-            plugin->file_size, file_status.st_size);
+            G_GINT64_FORMAT " != %"
+            G_GINT64_FORMAT, plugin->file_mtime, file_status.st_mtime,
+            (gint64) plugin->file_size, (gint64) file_status.st_size);
         gst_registry_remove_plugin (gst_registry_get_default (), plugin);
         newplugin = gst_plugin_load_file (filename, NULL);
         if (newplugin) {
@@ -797,6 +888,7 @@ gst_registry_scan_path_level (GstRegistry * registry, const gchar * path,
           newplugin->registered = TRUE;
           gst_object_unref (newplugin);
         }
+        changed = TRUE;
       }
       gst_object_unref (plugin);
 
@@ -806,6 +898,7 @@ gst_registry_scan_path_level (GstRegistry * registry, const gchar * path,
       if (newplugin) {
         newplugin->registered = TRUE;
         gst_object_unref (newplugin);
+        changed = TRUE;
       }
     }
 
@@ -813,6 +906,8 @@ gst_registry_scan_path_level (GstRegistry * registry, const gchar * path,
   }
 
   g_dir_close (dir);
+
+  return changed;
 }
 
 /**
@@ -823,20 +918,38 @@ gst_registry_scan_path_level (GstRegistry * registry, const gchar * path,
  * Add the given path to the registry. The syntax of the
  * path is specific to the registry. If the path has already been
  * added, do nothing.
+ *
+ * Returns: %TRUE if registry changed
  */
-void
+gboolean
 gst_registry_scan_path (GstRegistry * registry, const gchar * path)
 {
+  gboolean changed;
+
+  g_return_val_if_fail (GST_IS_REGISTRY (registry), FALSE);
+  g_return_val_if_fail (path != NULL, FALSE);
+
   GST_DEBUG_OBJECT (registry, "scanning path %s", path);
-  gst_registry_scan_path_level (registry, path, 10);
+  changed = gst_registry_scan_path_level (registry, path, 10);
+
+  GST_DEBUG_OBJECT (registry, "registry changed in path %s: %d", path, changed);
+
+  return changed;
 }
 
-void
-_gst_registry_remove_cache_plugins (GstRegistry * registry)
+/* Unref all plugins marked 'cached', to clear old plugins that no
+ * longer exist. Returns TRUE if any plugins were removed */
+gboolean
+_priv_gst_registry_remove_cache_plugins (GstRegistry * registry)
 {
   GList *g;
   GList *g_next;
   GstPlugin *plugin;
+  gboolean changed = FALSE;
+
+  g_return_val_if_fail (GST_IS_REGISTRY (registry), FALSE);
+
+  GST_OBJECT_LOCK (registry);
 
   GST_DEBUG_OBJECT (registry, "removing cached plugins");
   g = registry->plugins;
@@ -846,11 +959,17 @@ _gst_registry_remove_cache_plugins (GstRegistry * registry)
     if (plugin->flags & GST_PLUGIN_FLAG_CACHED) {
       GST_DEBUG_OBJECT (registry, "removing cached plugin \"%s\"",
           GST_STR_NULL (plugin->filename));
-      registry->plugins = g_list_remove (registry->plugins, plugin);
+      registry->plugins = g_list_delete_link (registry->plugins, g);
+      gst_registry_remove_features_for_plugin_unlocked (registry, plugin);
       gst_object_unref (plugin);
+      changed = TRUE;
     }
     g = g_next;
   }
+
+  GST_OBJECT_UNLOCK (registry);
+
+  return changed;
 }
 
 
@@ -874,18 +993,28 @@ GList *
 gst_registry_get_feature_list_by_plugin (GstRegistry * registry,
     const gchar * name)
 {
+  g_return_val_if_fail (GST_IS_REGISTRY (registry), NULL);
+  g_return_val_if_fail (name != NULL, NULL);
+
   return gst_registry_feature_filter (registry,
       _gst_plugin_feature_filter_plugin_name, FALSE, (gpointer) name);
 }
 
+/* Unref and delete the default registry */
 void
-_gst_registry_cleanup ()
+_priv_gst_registry_cleanup ()
 {
-  if (!_gst_registry_default)
-    return;
+  GstRegistry *registry;
 
-  gst_object_unref (_gst_registry_default);
-  _gst_registry_default = NULL;
+  g_static_mutex_lock (&_gst_registry_mutex);
+  if ((registry = _gst_registry_default) != NULL) {
+    _gst_registry_default = NULL;
+  }
+  g_static_mutex_unlock (&_gst_registry_mutex);
+
+  /* unref outside of the lock because we can. */
+  if (registry)
+    gst_object_unref (registry);
 }
 
 /**
