@@ -49,11 +49,13 @@
  * A segment structure is initialized with gst_segment_init(), which takes a #GstFormat
  * that will be used as the format of the segment values. The segment will be configured
  * with a start value of 0 and a stop/duration of -1, which is undefined. The default
- * rate is 1.0.
+ * rate and applied_rate is 1.0.
  *
  * If the segment is used for managing seeks, the segment duration should be set with
  * gst_segment_set_duration(). The public duration field contains the duration of the
- * segment.
+ * segment. When using the segment for seeking, the start and time members should 
+ * normally be left to their default 0 value. The stop position is left to -1 unless
+ * explicitly configured to a different value after a seek event.
  *
  * The current position in the segment should be set with the gst_segment_set_last_stop().
  * The public last_stop field contains the last set stop position in the segment.
@@ -64,20 +66,21 @@
  * If the cur_type was different from GST_SEEK_TYPE_NONE, playback continues from
  * the last_pos position, possibly with updated flags or rate.
  *
- * For elements that want to us #GstSegment to track the playback region, use
+ * For elements that want to use #GstSegment to track the playback region, use
  * gst_segment_set_newsegment() to update the segment fields with the information from
  * the newsegment event. The gst_segment_clip() method can be used to check and clip
  * the media data to the segment boundaries.
  *
  * For elements that want to synchronize to the pipeline clock, gst_segment_to_running_time()
  * can be used to convert a timestamp to a value that can be used to synchronize
- * to the clock. This function takes into account all accumulated segments.
+ * to the clock. This function takes into account all accumulated segments as well as
+ * any rate or applied_rate conversions.
  *
  * For elements that need to perform operations on media data in stream_time, 
  * gst_segment_to_stream_time() can be used to convert a timestamp and the segment
  * info to stream time (which is always between 0 and the duration of the stream).
  *
- * Last reviewed on 2005-12-12 (0.10.0)
+ * Last reviewed on 2006-05-03 (0.10.6)
  */
 
 static GstSegment *
@@ -89,7 +92,7 @@ gst_segment_copy (GstSegment * segment)
     result = gst_segment_new ();
     memcpy (result, segment, sizeof (GstSegment));
   }
-  return NULL;
+  return result;
 }
 
 GType
@@ -97,7 +100,7 @@ gst_segment_get_type (void)
 {
   static GType gst_segment_type = 0;
 
-  if (!gst_segment_type) {
+  if (G_UNLIKELY (gst_segment_type == 0)) {
     gst_segment_type = g_boxed_type_register_static ("GstSegment",
         (GBoxedCopyFunc) gst_segment_copy, (GBoxedFreeFunc) gst_segment_free);
   }
@@ -154,6 +157,7 @@ gst_segment_init (GstSegment * segment, GstFormat format)
 
   segment->rate = 1.0;
   segment->abs_rate = 1.0;
+  segment->applied_rate = 1.0;
   segment->format = format;
   segment->flags = 0;
   segment->start = 0;
@@ -168,13 +172,13 @@ gst_segment_init (GstSegment * segment, GstFormat format)
  * gst_segment_set_duration:
  * @segment: a #GstSegment structure.
  * @format: the format of the segment.
- * @duration: the duration of the segment info.
+ * @duration: the duration of the segment info or -1 if unknown.
  *
  * Set the duration of the segment to @duration. This function is mainly
  * used by elements that perform seeking and know the total duration of the
- * segment.
+ * segment. 
  * 
- * This field should be set to allow seeking request relative to the
+ * This field should be set to allow seeking requests relative to the
  * duration.
  */
 void
@@ -183,7 +187,7 @@ gst_segment_set_duration (GstSegment * segment, GstFormat format,
 {
   g_return_if_fail (segment != NULL);
 
-  if (segment->format == GST_FORMAT_UNDEFINED)
+  if (G_UNLIKELY (segment->format == GST_FORMAT_UNDEFINED))
     segment->format = format;
   else
     g_return_if_fail (segment->format == format);
@@ -198,6 +202,9 @@ gst_segment_set_duration (GstSegment * segment, GstFormat format,
  * @position: the position 
  *
  * Set the last observed stop position in the segment to @position.
+ *
+ * This field should be set to allow seeking requests relative to the
+ * current playing position.
  */
 void
 gst_segment_set_last_stop (GstSegment * segment, GstFormat format,
@@ -205,7 +212,7 @@ gst_segment_set_last_stop (GstSegment * segment, GstFormat format,
 {
   g_return_if_fail (segment != NULL);
 
-  if (segment->format == GST_FORMAT_UNDEFINED)
+  if (G_UNLIKELY (segment->format == GST_FORMAT_UNDEFINED))
     segment->format = format;
   else
     g_return_if_fail (segment->format == format);
@@ -219,69 +226,91 @@ gst_segment_set_last_stop (GstSegment * segment, GstFormat format,
  * @rate: the rate of the segment.
  * @format: the format of the segment.
  * @flags: the seek flags for the segment
- * @cur_type: the seek method
- * @cur: the seek start value
+ * @start_type: the seek method
+ * @start: the seek start value
  * @stop_type: the seek method
  * @stop: the seek stop value
- * @update: boolean holding whether an update the current segment is
- *    needed.
+ * @update: boolean holding whether last_stop was updated.
  *
- * Update the segment structure with the field values of a seek event.
+ * Update the segment structure with the field values of a seek event (see
+ * gst_event_new_seek()).
  *
- * After calling this method, the segment field last_stop will contain
- * the requested new position in the segment. If the cur_type is different
- * from GST_SEEK_TYPE_NONE, the current position is not updated and 
- * streaming should continue from the last position, possibly with
- * updated rate, flags or stop position.
+ * After calling this method, the segment field last_stop and time will
+ * contain the requested new position in the segment. The new requested
+ * position in the segment depends on @rate and @start_type and @stop_type. 
+ *
+ * For positive @rate, the new position in the segment is the new @segment
+ * start field when it was updated with a @start_type different from
+ * #GST_SEEK_TYPE_NONE. If no update was performed on @segment start position
+ * (#GST_SEEK_TYPE_NONE), @start is ignored and @segment last_stop is
+ * unmodified.
+ *
+ * For negative @rate, the new position in the segment is the new @segment
+ * stop field when it was updated with a @stop_type different from
+ * #GST_SEEK_TYPE_NONE. If no stop was previously configured in the segment, the
+ * duration of the segment will be used to update the stop position.
+ * If no update was performed on @segment stop position (#GST_SEEK_TYPE_NONE),
+ * @stop is ignored and @segment last_stop is unmodified.
+ *
+ * The applied rate of the segment will be set to 1.0 by default.
+ * If the caller can apply a rate change, it should update @segment
+ * rate and applied_rate after calling this function.
+ *
+ * @update will be set to TRUE if a seek should be performed to the segment 
+ * last_stop field. This field can be FALSE if, for example, only the @rate
+ * has been changed but not the playback position.
  */
 void
 gst_segment_set_seek (GstSegment * segment, gdouble rate,
     GstFormat format, GstSeekFlags flags,
-    GstSeekType cur_type, gint64 cur,
+    GstSeekType start_type, gint64 start,
     GstSeekType stop_type, gint64 stop, gboolean * update)
 {
   gboolean update_stop, update_start;
+  gint64 last_stop;
 
   g_return_if_fail (rate != 0.0);
   g_return_if_fail (segment != NULL);
 
-  if (segment->format == GST_FORMAT_UNDEFINED)
+  if (G_UNLIKELY (segment->format == GST_FORMAT_UNDEFINED))
     segment->format = format;
   else
     g_return_if_fail (segment->format == format);
 
-  update_stop = update_start = TRUE;
+  update_start = update_stop = TRUE;
 
   /* start is never invalid */
-  switch (cur_type) {
+  switch (start_type) {
     case GST_SEEK_TYPE_NONE:
       /* no update to segment */
-      cur = segment->start;
+      start = segment->start;
       update_start = FALSE;
       break;
     case GST_SEEK_TYPE_SET:
-      /* cur holds desired position */
+      /* start holds desired position, map -1 to the start */
+      if (start == -1)
+        start = 0;
       break;
     case GST_SEEK_TYPE_CUR:
-      /* add cur to currently configure segment */
-      cur = segment->start + cur;
+      /* add start to currently configure segment */
+      start = segment->start + start;
       break;
     case GST_SEEK_TYPE_END:
       if (segment->duration != -1) {
-        /* add cur to total length */
-        cur = segment->duration + cur;
+        /* add start to total length */
+        start = segment->duration + start;
       } else {
         /* no update if duration unknown */
-        cur = segment->start;
+        start = segment->start;
         update_start = FALSE;
       }
       break;
   }
   /* bring in sane range */
   if (segment->duration != -1)
-    cur = CLAMP (cur, 0, segment->duration);
+    start = CLAMP (start, 0, segment->duration);
   else
-    cur = MAX (cur, 0);
+    start = MAX (start, 0);
 
   /* stop can be -1 if we have not configured a stop. */
   switch (stop_type) {
@@ -290,7 +319,7 @@ gst_segment_set_seek (GstSegment * segment, gdouble rate,
       update_stop = FALSE;
       break;
     case GST_SEEK_TYPE_SET:
-      /* stop folds required value */
+      /* stop holds required value */
       break;
     case GST_SEEK_TYPE_CUR:
       if (segment->stop != -1)
@@ -318,20 +347,37 @@ gst_segment_set_seek (GstSegment * segment, gdouble rate,
 
   /* we can't have stop before start */
   if (stop != -1)
-    g_return_if_fail (cur <= stop);
+    g_return_if_fail (start <= stop);
 
   segment->rate = rate;
   segment->abs_rate = ABS (rate);
+  segment->applied_rate = 1.0;
   segment->flags = flags;
-  segment->start = cur;
-  if (update_start) {
-    segment->last_stop = cur;
+  segment->start = start;
+  last_stop = segment->last_stop;
+  if (update_start && rate > 0.0) {
+    last_stop = start;
   }
-  segment->time = segment->last_stop;
-  segment->stop = stop;
-
+  if (update_stop && rate < 0.0) {
+    if (stop != -1)
+      last_stop = stop;
+    else {
+      if (segment->duration != -1)
+        last_stop = segment->duration;
+      else
+        last_stop = 0;
+    }
+  }
+  /* set update arg to reflect update of last_stop */
   if (update)
-    *update = update_start || update_stop;
+    *update = last_stop != segment->last_stop;
+
+  /* update new position */
+  if (last_stop != segment->last_stop)
+    segment->last_stop = last_stop;
+
+  segment->time = start;
+  segment->stop = stop;
 }
 
 /**
@@ -344,18 +390,44 @@ gst_segment_set_seek (GstSegment * segment, gdouble rate,
  * @stop: the new stop value
  * @time: the new stream time
  *
- * Update the segment structure with the field values of a new segment event.
+ * Update the segment structure with the field values of a new segment event and
+ * with a default applied_rate of 1.0.
+ *
+ * Since: 0.10.6
  */
 void
 gst_segment_set_newsegment (GstSegment * segment, gboolean update, gdouble rate,
     GstFormat format, gint64 start, gint64 stop, gint64 time)
+{
+  gst_segment_set_newsegment_full (segment, update, rate, 1.0, format, start,
+      stop, time);
+}
+
+/**
+ * gst_segment_set_newsegment_full:
+ * @segment: a #GstSegment structure.
+ * @update: flag indicating a new segment is started or updated
+ * @rate: the rate of the segment.
+ * @applied_rate: the applied rate of the segment.
+ * @format: the format of the segment.
+ * @start: the new start value
+ * @stop: the new stop value
+ * @time: the new stream time
+ *
+ * Update the segment structure with the field values of a new segment event.
+ */
+void
+gst_segment_set_newsegment_full (GstSegment * segment, gboolean update,
+    gdouble rate, gdouble applied_rate, GstFormat format, gint64 start,
+    gint64 stop, gint64 time)
 {
   gint64 duration;
 
   g_return_if_fail (rate != 0.0);
+  g_return_if_fail (applied_rate != 0.0);
   g_return_if_fail (segment != NULL);
 
-  if (segment->format == GST_FORMAT_UNDEFINED)
+  if (G_UNLIKELY (segment->format == GST_FORMAT_UNDEFINED))
     segment->format = format;
 
   /* any other format with 0 also gives time 0, the other values are
@@ -373,30 +445,39 @@ gst_segment_set_newsegment (GstSegment * segment, gboolean update, gdouble rate,
   if (update) {
     /* an update to the current segment is done, elapsed time is
      * difference between the old start and new start. */
-    duration = start - segment->start;
+    if (start > segment->start)
+      duration = start - segment->start;
+    else
+      duration = 0;
   } else {
     /* the new segment has to be aligned with the old segment.
      * We first update the accumulated time of the previous
      * segment. the accumulated time is used when syncing to the
      * clock. 
      */
-    if (GST_CLOCK_TIME_IS_VALID (segment->stop)) {
+    if (segment->stop != -1) {
       duration = segment->stop - segment->start;
-    } else if (GST_CLOCK_TIME_IS_VALID (segment->last_stop)) {
+    } else if (segment->last_stop != -1) {
       /* else use last seen timestamp as segment stop */
       duration = segment->last_stop - segment->start;
     } else {
-      /* else we don't know */
+      /* else we don't know and throw a warning.. really, this should
+       * be fixed in the element. */
+      g_warning ("closing segment of unknown duration, assuming duration of 0");
       duration = 0;
     }
   }
   /* use previous rate to calculate duration */
-  segment->accum += gst_gdouble_to_guint64 (
-      (gst_guint64_to_gdouble (duration) / segment->abs_rate));
+  if (segment->abs_rate != 1.0)
+    duration /= segment->abs_rate;
+
+  /* accumulate duration */
+  segment->accum += duration;
 
   /* then update the current segment */
   segment->rate = rate;
   segment->abs_rate = ABS (rate);
+  segment->applied_rate = applied_rate;
   segment->start = start;
   segment->last_stop = start;
   segment->stop = stop;
@@ -410,33 +491,69 @@ gst_segment_set_newsegment (GstSegment * segment, gboolean update, gdouble rate,
  * @position: the position in the segment
  *
  * Translate @position to stream time using the currently configured 
- * segment.
+ * segment. The @position value must be between @segment start and
+ * stop value. 
  *
  * This function is typically used by elements that need to operate on
  * the stream time of the buffers it receives, such as effect plugins.
+ * In those use cases, @position is typically the buffer timestamp or 
+ * clock time that one wants to convert to the stream time.
+ * The stream time is always between 0 and the total duration of the 
+ * media stream. 
  *
- * Returns: the position in stream_time.
+ * Returns: the position in stream_time or -1 when an invalid position
+ * was given.
  */
 gint64
 gst_segment_to_stream_time (GstSegment * segment, GstFormat format,
     gint64 position)
 {
-  gint64 result, time;
+  gint64 result;
+  gdouble abs_applied_rate;
 
-  g_return_val_if_fail (segment != NULL, FALSE);
+  g_return_val_if_fail (segment != NULL, -1);
 
-  if (segment->format == GST_FORMAT_UNDEFINED)
+  /* format does not matter for -1 */
+  if (G_UNLIKELY (position == -1))
+    return -1;
+
+  if (G_UNLIKELY (segment->format == GST_FORMAT_UNDEFINED))
     segment->format = format;
   else
-    g_return_val_if_fail (segment->format == format, FALSE);
+    g_return_val_if_fail (segment->format == format, -1);
 
-  if ((time = segment->time) == -1)
-    time = 0;
+  /* outside of the segment boundary stop */
+  if (G_UNLIKELY (segment->stop != -1 && position > segment->stop))
+    return -1;
 
-  if (position != -1 && position >= segment->start)
-    result = ((position - segment->start) / segment->abs_rate) + time;
-  else
-    result = -1;
+  /* before the segment boundary */
+  if (G_UNLIKELY (position < segment->start))
+    return -1;
+
+  /* time must be known */
+  if (G_UNLIKELY (segment->time == -1))
+    return -1;
+
+  /* bring to uncorrected position in segment */
+  result = position - segment->start;
+
+  abs_applied_rate = ABS (segment->applied_rate);
+
+  /* correct for applied rate if needed */
+  if (abs_applied_rate != 1.0)
+    result *= abs_applied_rate;
+
+  /* add or subtract from segment time based on applied rate */
+  if (segment->applied_rate > 0.0) {
+    /* correct for segment time */
+    result += segment->time;
+  } else {
+    /* correct for segment time, clamp at 0 */
+    if (segment->time > result)
+      result = segment->time - result;
+    else
+      result = 0;
+  }
 
   return result;
 }
@@ -448,12 +565,18 @@ gst_segment_to_stream_time (GstSegment * segment, GstFormat format,
  * @position: the position in the segment
  *
  * Translate @position to the total running time using the currently configured 
- * segment.
+ * and previously accumulated segments. Position is a value between @segment
+ * start and stop time.
  *
  * This function is typically used by elements that need to synchronize to the
- * global clock in a pipeline.
+ * global clock in a pipeline. The runnning time is a constantly increasing value
+ * starting from 0. When gst_segment_init() is called, this value will reset to
+ * 0.
  *
- * Returns: the position as the total running time.
+ * This function returns -1 if the position is outside of @segment start and stop.
+ *
+ * Returns: the position as the total running time or -1 when an invalid position
+ * was given.
  */
 gint64
 gst_segment_to_running_time (GstSegment * segment, GstFormat format,
@@ -463,15 +586,42 @@ gst_segment_to_running_time (GstSegment * segment, GstFormat format,
 
   g_return_val_if_fail (segment != NULL, -1);
 
-  if (segment->format == GST_FORMAT_UNDEFINED)
+  if (G_UNLIKELY (position == -1))
+    return -1;
+
+  if (G_UNLIKELY (segment->format == GST_FORMAT_UNDEFINED))
     segment->format = format;
   else if (segment->accum)
     g_return_val_if_fail (segment->format == format, -1);
 
-  if (position != -1 && position >= segment->start)
-    result = ((position - segment->start) / segment->abs_rate) + segment->accum;
-  else
-    result = -1;
+  /* before the segment boundary */
+  if (G_UNLIKELY (position < segment->start))
+    return -1;
+
+  if (segment->rate > 0.0) {
+    /* outside of the segment boundary stop */
+    if (G_UNLIKELY (segment->stop != -1 && position > segment->stop))
+      return -1;
+
+    /* bring to uncorrected position in segment */
+    result = position - segment->start;
+  } else {
+    /* cannot continue if no stop position set or outside of
+     * the segment. */
+    if (G_UNLIKELY (segment->stop == -1 || position > segment->stop))
+      return -1;
+
+    /* bring to uncorrected position in segment */
+    result = segment->stop - position;
+  }
+
+  /* scale based on the rate, avoid division by and conversion to 
+   * float when not needed */
+  if (segment->abs_rate != 1.0)
+    result /= segment->abs_rate;
+
+  /* correct for accumulated segments */
+  result += segment->accum;
 
   return result;
 }
@@ -486,10 +636,19 @@ gst_segment_to_running_time (GstSegment * segment, GstFormat format,
  * @clip_stop: the clipped stop position in the segment
  *
  * Clip the given @start and @stop values to the segment boundaries given
- * in @segment.
+ * in @segment. @start and @stop are compared and clipped to @segment 
+ * start and stop values.
  *
- * Returns: TRUE if the given @start and @stop times fall partially in 
- *     @segment, FALSE if the values are completely outside of the segment.
+ * If the function returns FALSE, @start and @stop are known to fall
+ * outside of @segment and @clip_start and @clip_stop are not updated.
+ *
+ * When the function returns TRUE, @clip_start and @clip_stop will be
+ * updated. If @clip_start or @clip_stop are different from @start or @stop
+ * respectively, the region fell partially in the segment.
+ *
+ * Returns: TRUE if the given @start and @stop times fall partially or 
+ *     completely in @segment, FALSE if the values are completely outside 
+ *     of the segment.
  */
 gboolean
 gst_segment_clip (GstSegment * segment, GstFormat format, gint64 start,
@@ -497,27 +656,27 @@ gst_segment_clip (GstSegment * segment, GstFormat format, gint64 start,
 {
   g_return_val_if_fail (segment != NULL, FALSE);
 
-  if (segment->format == GST_FORMAT_UNDEFINED)
+  if (G_UNLIKELY (segment->format == GST_FORMAT_UNDEFINED))
     segment->format = format;
   else
     g_return_val_if_fail (segment->format == format, FALSE);
 
-  /* we need a valid start position */
-  if (start == -1)
-    return FALSE;
-
-  /* if we have a stop position and start is bigger, we're
-   * outside of the segment */
-  if (segment->stop != -1 && start >= segment->stop)
+  /* if we have a stop position and a valid start and start is bigger, 
+   * we're outside of the segment */
+  if (G_UNLIKELY (segment->stop != -1 && start != -1 && start >= segment->stop))
     return FALSE;
 
   /* if a stop position is given and is before the segment start,
    * we're outside of the segment */
-  if (stop != -1 && stop <= segment->start)
+  if (G_UNLIKELY (stop != -1 && stop != start && stop <= segment->start))
     return FALSE;
 
-  if (clip_start)
-    *clip_start = MAX (start, segment->start);
+  if (clip_start) {
+    if (start == -1)
+      *clip_start = -1;
+    else
+      *clip_start = MAX (start, segment->start);
+  }
 
   if (clip_stop) {
     if (stop == -1)
