@@ -39,6 +39,10 @@ GST_START_TEST (test_caps)
   fail_unless (GST_BUFFER_CAPS (buffer) == caps);
   ASSERT_CAPS_REFCOUNT (caps, "caps", 2);
 
+  fail_unless (gst_buffer_get_caps (buffer) == caps);
+  gst_caps_unref (caps);
+  ASSERT_CAPS_REFCOUNT (caps, "caps", 2);
+
   caps2 = gst_caps_from_string ("audio/x-raw-float");
   ASSERT_CAPS_REFCOUNT (caps2, "caps2", 1);
 
@@ -68,6 +72,11 @@ GST_START_TEST (test_subbuffer)
 
   buffer = gst_buffer_new_and_alloc (4);
   memset (GST_BUFFER_DATA (buffer), 0, 4);
+  /* set some metadata */
+  GST_BUFFER_TIMESTAMP (buffer) = 1;
+  GST_BUFFER_DURATION (buffer) = 2;
+  GST_BUFFER_OFFSET (buffer) = 3;
+  GST_BUFFER_OFFSET_END (buffer) = 4;
 
   sub = gst_buffer_create_sub (buffer, 1, 2);
   fail_if (sub == NULL, "create_sub of buffer returned NULL");
@@ -76,6 +85,48 @@ GST_START_TEST (test_subbuffer)
           2) == 0, "subbuffer contains the wrong data");
   ASSERT_BUFFER_REFCOUNT (buffer, "parent", 2);
   ASSERT_BUFFER_REFCOUNT (sub, "subbuffer", 1);
+  fail_unless (GST_BUFFER_TIMESTAMP (sub) == -1,
+      "subbuffer has wrong timestamp");
+  fail_unless (GST_BUFFER_DURATION (sub) == -1, "subbuffer has wrong duration");
+  fail_unless (GST_BUFFER_OFFSET (sub) == -1, "subbuffer has wrong offset");
+  fail_unless (GST_BUFFER_OFFSET_END (sub) == -1,
+      "subbuffer has wrong offset end");
+  gst_buffer_unref (sub);
+
+  /* create a subbuffer of size 0 */
+  sub = gst_buffer_create_sub (buffer, 1, 0);
+  fail_if (sub == NULL, "create_sub of buffer returned NULL");
+  fail_unless (GST_BUFFER_SIZE (sub) == 0, "subbuffer has wrong size");
+  fail_unless (memcmp (GST_BUFFER_DATA (buffer) + 1, GST_BUFFER_DATA (sub),
+          0) == 0, "subbuffer contains the wrong data");
+  ASSERT_BUFFER_REFCOUNT (buffer, "parent", 2);
+  ASSERT_BUFFER_REFCOUNT (sub, "subbuffer", 1);
+  gst_buffer_unref (sub);
+
+  /* test if metadata is coppied, not a complete buffer copy so only the
+   * timestamp and offset fields are copied. */
+  sub = gst_buffer_create_sub (buffer, 0, 1);
+  fail_if (sub == NULL, "create_sub of buffer returned NULL");
+  fail_unless (GST_BUFFER_SIZE (sub) == 1, "subbuffer has wrong size");
+  fail_unless (GST_BUFFER_TIMESTAMP (sub) == 1,
+      "subbuffer has wrong timestamp");
+  fail_unless (GST_BUFFER_OFFSET (sub) == 3, "subbuffer has wrong offset");
+  fail_unless (GST_BUFFER_DURATION (sub) == -1, "subbuffer has wrong duration");
+  fail_unless (GST_BUFFER_OFFSET_END (sub) == -1,
+      "subbuffer has wrong offset end");
+  gst_buffer_unref (sub);
+
+  /* test if metadata is coppied, a complete buffer is copied so all the timing
+   * fields should be copied. */
+  sub = gst_buffer_create_sub (buffer, 0, 4);
+  fail_if (sub == NULL, "create_sub of buffer returned NULL");
+  fail_unless (GST_BUFFER_SIZE (sub) == 4, "subbuffer has wrong size");
+  fail_unless (GST_BUFFER_TIMESTAMP (sub) == 1,
+      "subbuffer has wrong timestamp");
+  fail_unless (GST_BUFFER_DURATION (sub) == 2, "subbuffer has wrong duration");
+  fail_unless (GST_BUFFER_OFFSET (sub) == 3, "subbuffer has wrong offset");
+  fail_unless (GST_BUFFER_OFFSET_END (sub) == 4,
+      "subbuffer has wrong offset end");
 
   /* clean up */
   gst_buffer_unref (sub);
@@ -182,7 +233,7 @@ GST_START_TEST (test_span)
   gst_buffer_unref (span);
   ASSERT_BUFFER_REFCOUNT (buffer, "parent", 3);
 
-/* clean up */
+  /* clean up */
   gst_buffer_unref (sub1);
   gst_buffer_unref (sub2);
   gst_buffer_unref (buffer);
@@ -317,8 +368,41 @@ GST_START_TEST (test_metadata_writable)
 
 GST_END_TEST;
 
+GST_START_TEST (test_copy)
+{
+  GstBuffer *buffer, *copy;
+
+  buffer = gst_buffer_new_and_alloc (4);
+  ASSERT_BUFFER_REFCOUNT (buffer, "buffer", 1);
+
+  copy = gst_buffer_copy (buffer);
+  ASSERT_BUFFER_REFCOUNT (buffer, "buffer", 1);
+  ASSERT_BUFFER_REFCOUNT (copy, "copy", 1);
+  /* data must be copied and thus point to different memory */
+  fail_if (GST_BUFFER_DATA (buffer) == GST_BUFFER_DATA (copy));
+
+  gst_buffer_unref (copy);
+  gst_buffer_unref (buffer);
+
+  /* a 0-sized buffer has NULL data as per docs */
+  buffer = gst_buffer_new_and_alloc (0);
+  fail_unless (GST_BUFFER_DATA (buffer) == NULL);
+  fail_unless (GST_BUFFER_SIZE (buffer) == 0);
+
+  /* copying a 0-sized buffer should not crash and also set
+   * the data member NULL. */
+  copy = gst_buffer_copy (buffer);
+  fail_unless (GST_BUFFER_DATA (copy) == NULL);
+  fail_unless (GST_BUFFER_SIZE (copy) == 0);
+
+  gst_buffer_unref (copy);
+  gst_buffer_unref (buffer);
+}
+
+GST_END_TEST;
+
 Suite *
-gst_test_suite (void)
+gst_buffer_suite (void)
 {
   Suite *s = suite_create ("GstBuffer");
   TCase *tc_chain = tcase_create ("general");
@@ -331,22 +415,9 @@ gst_test_suite (void)
   tcase_add_test (tc_chain, test_is_span_fast);
   tcase_add_test (tc_chain, test_span);
   tcase_add_test (tc_chain, test_metadata_writable);
+  tcase_add_test (tc_chain, test_copy);
+
   return s;
 }
 
-int
-main (int argc, char **argv)
-{
-  int nf;
-
-  Suite *s = gst_test_suite ();
-  SRunner *sr = srunner_create (s);
-
-  gst_check_init (&argc, &argv);
-
-  srunner_run_all (sr, CK_NORMAL);
-  nf = srunner_ntests_failed (sr);
-  srunner_free (sr);
-
-  return nf;
-}
+GST_CHECK_MAIN (gst_buffer);
